/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"

#include <jack/jack.h>
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef HAVE_LADCCA
#include <ladcca/ladcca.h>
extern cca_client_t * cca_client;
#endif

#include "process.h"
#include "control-message.h"
#include "lock-free-fifo.h"
#include "chain.h"
#include "xmalloc.h"

jack_nframes_t sample_rate;
jack_nframes_t buffer_size;

int process_control_messages (process_info_t * procinfo) {
  static int quitting = 0;
  ctrlmsg_t ctrlmsg;
  int err = 0;
  ladspa_chain_link_t * chain_link;
  
  if (quitting) return quitting;
  
  while (lff_read (procinfo->gui_to_process, &ctrlmsg) == 0) {
  
    switch (ctrlmsg.type) {
    
      /* add a link to the end of the plugin chain */
      case CTRLMSG_ADD:
        add_plugin (procinfo, (ladspa_chain_link_t *) ctrlmsg.pointer);
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      /* remove a link (plugin will be sent back) */
      case CTRLMSG_REMOVE:
        chain_link = remove_plugin (procinfo, ctrlmsg.number);
        ctrlmsg.pointer = chain_link;
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      /* enable/disable a plugin */
      case CTRLMSG_ABLE:
        ablise_plugin (procinfo, ctrlmsg.number,
                             GPOINTER_TO_INT(ctrlmsg.pointer));
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
      
      /* move a plugin up or down */  
      case CTRLMSG_MOVE:
/*        printf("%s: moving plugin %d %s\n", __FUNCTION__,
               ctrlmsg.number, (GPOINTER_TO_INT(ctrlmsg.pointer)) ? "up" : "down");*/
        move_plugin (procinfo, ctrlmsg.number,
                           GPOINTER_TO_INT(ctrlmsg.pointer));
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
              
      /* change an existing plugin for a new one (existing plugin
         will be sent back) */
      case CTRLMSG_CHANGE:
        chain_link = change_plugin (procinfo, ctrlmsg.number,
                             (ladspa_chain_link_t *) ctrlmsg.pointer);
        ctrlmsg.pointer = chain_link;
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      case CTRLMSG_QUIT:
        quitting = 1;
        return 1;
        break;
    }
    
    if (err) {
      fprintf (stderr, "%s: gui fifo is out of space\n", __FUNCTION__);
      return err;
    }
  }
  return 0;
}

/** process messages for plugins' control ports */
void process_control_port_messages (process_info_t * procinfo) {
  ladspa_chain_link_t * plugin;
  unsigned long i;
  
  if (!procinfo->chain) return;
  
  for (plugin = procinfo->chain; plugin; plugin = plugin->next) {
    if (plugin->control_port_count > 0) {
      if (plugin->mono) {
        for (i = 0; i < plugin->control_port_count; i++) {
          while (lff_read (plugin->holder.mono_pair.left->control_fifos + i,
                           plugin->holder.mono_pair.left->control_memory + i) == 0) {}
          while (lff_read (plugin->holder.mono_pair.right->control_fifos + i,
                           plugin->holder.mono_pair.right->control_memory + i) == 0) {}
        }
      } else {
        for (i = 0; i < plugin->control_port_count; i++) {
          while (lff_read (plugin->holder.stereo->control_fifos + i,
                           plugin->holder.stereo->control_memory + i) == 0) {}
        }
      }
    }
  }
}

int get_jack_buffers (process_info_t * procinfo, jack_nframes_t frames) {
  procinfo->jack_buffers[IN_L] = jack_port_get_buffer (procinfo->jack_ports[IN_L], frames);
  if (!procinfo->jack_buffers[IN_L]) {
    fprintf (stderr, "%s: no jack buffer for left input port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[IN_R] = jack_port_get_buffer (procinfo->jack_ports[IN_R], frames);
  if (!procinfo->jack_buffers[IN_R]) {
    fprintf (stderr, "%s: no jack buffer for right input port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[OUT_L] = jack_port_get_buffer (procinfo->jack_ports[OUT_L], frames);
  if (!procinfo->jack_buffers[OUT_L]) {
    fprintf (stderr, "%s: no jack buffer for left output port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[OUT_R] = jack_port_get_buffer (procinfo->jack_ports[OUT_R], frames);
  if (!procinfo->jack_buffers[OUT_R]) {
    fprintf (stderr, "%s: no jack buffer for right output port\n", __FUNCTION__);
    return 1;
  }
  
  return 0;
}

ladspa_chain_link_t * get_first_enabled_plugin (process_info_t * procinfo) {
  ladspa_chain_link_t * first_enabled;
  
  if (!procinfo->chain) return NULL;

  for (first_enabled = procinfo->chain;
       first_enabled;
       first_enabled = first_enabled->next) {
    if (first_enabled->enabled) return first_enabled;
  }
 
  return NULL;
}

ladspa_chain_link_t * get_last_enabled_plugin (process_info_t * procinfo) {
  ladspa_chain_link_t * last_enabled;
  
  if (!procinfo->chain) return NULL;

  for (last_enabled = procinfo->chain_end;
       last_enabled;
       last_enabled = last_enabled->prev) {
    if (last_enabled->enabled) return last_enabled;
  }
  
  return NULL;
}

void connect_chain (process_info_t * procinfo, jack_nframes_t frames) {
  ladspa_chain_link_t * first_enabled, * last_enabled, * plugin;
  
  if (!procinfo->chain) return;
  
  first_enabled = get_first_enabled_plugin (procinfo);
  if (!first_enabled) return;
  
  last_enabled = get_last_enabled_plugin (procinfo);

  /* ensure that all the of the enabled plugins are connected to their memory */
  if (first_enabled != last_enabled) {
    connect_output_ports (first_enabled);
    connect_input_ports (last_enabled);
    for (plugin = first_enabled->next;
         plugin && (plugin != last_enabled);
         plugin = plugin->next) {
      connect_input_ports (plugin);
      connect_output_ports (plugin);
    }
  }

  /* input buffers for first plugin */
  if (first_enabled->mono) {
    first_enabled->descriptor->
      connect_port (first_enabled->holder.mono_pair.left->instance,
                    first_enabled->port_indicies[IN_L],
                    procinfo->jack_buffers[IN_L]);
    first_enabled->descriptor->
      connect_port (first_enabled->holder.mono_pair.right->instance,
                    first_enabled->port_indicies[IN_R],
                    procinfo->jack_buffers[IN_R]);
  } else {
    first_enabled->descriptor->
      connect_port (first_enabled->holder.stereo->instance,
                    first_enabled->port_indicies[IN_L],
                    procinfo->jack_buffers[IN_L]);
    first_enabled->descriptor->
      connect_port (first_enabled->holder.stereo->instance,
                    first_enabled->port_indicies[IN_R],
                    procinfo->jack_buffers[IN_R]);
  }
  
  
  /* output buffers for last plugin */
  if (last_enabled->mono) {
    last_enabled->descriptor->
      connect_port (last_enabled->holder.mono_pair.left->instance,
                    last_enabled->port_indicies[OUT_L],
                    procinfo->jack_buffers[OUT_L]);
    last_enabled->descriptor->
      connect_port (last_enabled->holder.mono_pair.right->instance,
                    last_enabled->port_indicies[OUT_R],
                    procinfo->jack_buffers[OUT_R]);
  } else {
    last_enabled->descriptor->
      connect_port (last_enabled->holder.stereo->instance,
                    last_enabled->port_indicies[OUT_L],
                    procinfo->jack_buffers[OUT_L]);
    last_enabled->descriptor->
      connect_port (last_enabled->holder.stereo->instance,
                    last_enabled->port_indicies[OUT_R],
                    procinfo->jack_buffers[OUT_R]);
  }
  

}

void process_chain (process_info_t * procinfo, jack_nframes_t frames) {
  ladspa_chain_link_t * first_enabled, * last_enabled = 0, * plugin;

  first_enabled = get_first_enabled_plugin (procinfo);
  
  /* no chain; just copy input to output */
  if (!procinfo->chain || !first_enabled) {
    memcpy (procinfo->jack_buffers[OUT_L], procinfo->jack_buffers[IN_L], sizeof(LADSPA_Data) * frames);
    memcpy (procinfo->jack_buffers[OUT_R], procinfo->jack_buffers[IN_R], sizeof(LADSPA_Data) * frames);
    return;
  }
  
  /* all past here is guaranteed to have at least 1 enabled plugin */

  last_enabled = get_last_enabled_plugin (procinfo);
  
  for (plugin = first_enabled;
       plugin;
       plugin = plugin->next) {
    if (plugin->enabled) {
    
      if (plugin->mono) {
        plugin->descriptor->
          run (plugin->holder.mono_pair.left->instance, frames);
        plugin->descriptor->
          run (plugin->holder.mono_pair.right->instance, frames);
      } else {
        plugin->descriptor->
          run (plugin->holder.stereo->instance, frames);
      }
      
      if (plugin == last_enabled) break;

    } else {
    
      /* copy the data through */
      memcpy (plugin->audio_memory[0],
              plugin->prev->audio_memory[0],
              sizeof(LADSPA_Data) * frames);
      memcpy (plugin->audio_memory[1],
              plugin->prev->audio_memory[1],
              sizeof(LADSPA_Data) * frames);
    }
  }
}

int process (jack_nframes_t frames, void * data) {
  int err, quitting;
  process_info_t * procinfo;
  
  procinfo = (process_info_t *) data;
  
  if (!procinfo) {
    fprintf (stderr, "%s: no process_info from jack!\n", __FUNCTION__);
    return 1;
  }
  
  quitting = process_control_messages (procinfo);
  
  if (quitting) return 1;
  
  process_control_port_messages (procinfo);
  
  err = get_jack_buffers (procinfo, frames);
  if (err) {
    fprintf(stderr, "%s: failed to get jack ports, not processing\n", __FUNCTION__);
    return 1;
  }
  
  connect_chain (procinfo, frames);
  
  process_chain (procinfo, frames);
  
  return 0;
}



/*******************************************
 ************** non RT stuff ***************
 *******************************************/

int init_jack (process_info_t * process_info,
               const char * client_name) {
  int i;
  struct { const char * name;
           int index;
           enum JackPortFlags flags;
         } port_map[] = { { "in_1", IN_R, JackPortIsInput },
                          { "in_2", IN_L, JackPortIsInput },
                          { "out_1", OUT_R, JackPortIsOutput },
                          { "out_2", OUT_L, JackPortIsOutput } };
           
  process_info->jack_client = jack_client_new (client_name);
  
  if (!process_info->jack_client) {
    fprintf (stderr, "%s: could not create jack client; exiting\n", __FUNCTION__);
    fprintf (stderr, "%s:   is the jack server running?\n", __FUNCTION__);
    exit (1);
  }
  
  for (i = 0; i < 4; i++) {
    process_info->jack_ports[port_map[i].index] =
      jack_port_register (process_info->jack_client,
                          /* port name */    port_map[i].name,
                          /* port type */    JACK_DEFAULT_AUDIO_TYPE,
                          /* flags */        port_map[i].flags,
                          /* buffer size */  0);
    
    if (!process_info->jack_ports[port_map[i].index]) {
      fprintf (stderr, "%s: could not register port '%s'; exiting\n",
               __FUNCTION__, port_map[i].name);
      exit (1);
    }
  }
  
  sample_rate = jack_get_sample_rate (process_info->jack_client);
  buffer_size = jack_get_sample_rate (process_info->jack_client);
  
  jack_set_process_callback (process_info->jack_client, process, process_info);
  
  /* init the fifos */
  process_info->gui_to_process = xmalloc (sizeof (lff_t)); 
  process_info->process_to_gui = xmalloc (sizeof (lff_t)); 
  lff_init (process_info->gui_to_process, 32, sizeof (ctrlmsg_t));
  lff_init (process_info->process_to_gui, 32, sizeof (ctrlmsg_t));
  
  process_info->chain = NULL;
  process_info->chain_end = NULL;

#ifdef HAVE_LADCCA
  cca_jack_client_name (cca_client, client_name);
#endif

  return 0;
}

void destroy_process_info (process_info_t * process_info) {
  ladspa_chain_link_t * chain_link, * next_chain_link;
  
  for (chain_link = process_info->chain_end;
       chain_link;
       chain_link = next_chain_link) {
    next_chain_link = chain_link->next;
    
    destroy_chain_link (chain_link);
  }
  
  lff_free (process_info->gui_to_process);
  lff_free (process_info->process_to_gui);
  free (process_info->gui_to_process);
  free (process_info->process_to_gui);
}

